/* For pthreads it may be important on some implementations to ensure
   that pthread.h is included before anything else */
#include <pthread.h>

#include <stdio.h>
#include <stdlib.h>
#include "workers.h"

#define MAX_NUM_WORKERS 128

static int numWorkers = 0;
static int numJobs = 0;
static void (*work)(int i, void *data);
static void *data;
static void (*signal_error)(const char *msg) = NULL;
static int numProcs = 64;  /**** need to try to determine this at runtime */

/**** should be able to configure these; maybe have several strategies */
/* Variables to control spin-wait in semaphores */
static long long standard_spin_count = 100000LL;
static long long reduced_spin_count = 1000LL;
static long long spin_count = 100000LL;
static int spin_adjust_interval = 200;
static int spin_reduce = 4;

/* Forward declarations for implementation-dependent routines */
static void releaseWorkers(int njobs, void (*workfun)(int i, void *data),
			   void *workdata);
static void waitForWorkers(void);
static int joinWorkers(void);
static void exit_work(int i, void *data);
static int startWorkerThreads(int n);


/* Error handling routine. Must not return normally. */
static void error(const char *msg)
{
    if (signal_error != NULL)
	signal_error(msg);
    else fprintf(stderr, "%s\n", msg);
    abort();
}


/* Utility Function */
void getLoopChunkRange(int k, size_t n, int P, size_t *start, size_t *end)
{
    size_t N = n / P;
    size_t R = n % P;
    *start =  k < R ? k * (N + 1) : R + k * N;
    *end = k < R ? (k + 1) * (N + 1) : R + (k + 1) * N;
}
 
/* The implementations of the following public functions should be
   independent of the particular thread system used. */

void initWorkers(int n, void (*err)(const char *msg))
{
    signal_error = err;

    if (n < 1) error("too few workers");
    if (n > MAX_NUM_WORKERS) error("too many workers");

    if (numWorkers > 0) {
	if (numWorkers == n) return;
	else stopWorkers();
    }
    numWorkers = startWorkerThreads(n);
    if (numWorkers != n) {
	numWorkers = 0;
	/**** shut down any running threads at this point? */
	/**** not sure if it is even possible to get here */
	error("one of the thread creations failed");
    }
}

/**** allow more jobs than workers? */
void runWorkers(int njobs, void (*workfun)(int i, void *data), void *workdata)
{
    if (numWorkers < 1) error("workers have not been initialized");

    if (njobs <= numProcs) {
	static int spin_adjust = 0;
	if (-- spin_adjust < 0) {
	    spin_adjust = spin_adjust_interval;
	    spin_count = standard_spin_count;
	}
    }
    else
	spin_count = reduced_spin_count;

    if (numWorkers == 1) {
	int k;
	for (k = 0; k < njobs; k++)
	    workfun(k, workdata);
    }
    else if (njobs == 1) 
	workfun(0, workdata);
    else {
	releaseWorkers(njobs - 1, workfun, workdata);
	workfun(njobs - 1, workdata);
	waitForWorkers();
    }
}

void stopWorkers()
{
    int join_failed = 0;

    releaseWorkers(numWorkers, exit_work, NULL);
    join_failed = joinWorkers();
    numWorkers = 0;
    if (join_failed) error("a thread join failed");
}


/* The following data structures and routines are pthread-dependent
   and/or OS dependent */
/**** Should do more checking of pthread_foo status values. */

/* Atomic operations on integers */
#ifdef __APPLE__
#include <libkern/OSAtomic.h>

static int atomic_compare_and_swap(int *val, int target, int new)
{
    return OSAtomicCompareAndSwapIntBarrier(target, new, val);
}

static int atomic_decrement(int *val)
{
    return OSAtomicDecrement32Barrier(val);
}
#elif defined( __WIN32__)
#include <windows.h>

static int atomic_compare_and_swap(int *val, int target, int new)
{
    return InterlockedCompareExchange((LONG volatile *) val, new, target);
}

static int atomic_decrement(int *val)
{
    return InterlockedDecrement((LONG volatile *)val);
}
#elif defined(__sun__)
/* for Solaris; may also work on NetBSD */
#include <atomic.h>
static int atomic_compare_and_swap(int *val, int target, int new)
{
    return (int) atomic_cas_uint((volatile uint_t *) val, (uint_t) target,
				 (uint_t) new);
}

static int atomic_decrement(int *val)
{
    atomic_dec_uint((volatile uint_t *) val);
    return *val;
}
#elif ! defined(NO_SYNC_FUNS)
#include <sys/syscall.h>

static int atomic_compare_and_swap(int *val, int target, int new)
{
    return __sync_val_compare_and_swap (val, target, new);
}

static int atomic_decrement(int *val)
{
    return __sync_sub_and_fetch(val, 1);
}
#else
static int atomic_decrement(int *val)
{
    static pthread_mutex_t dec_mtx = PTHREAD_MUTEX_INITIALIZER;
    int new;
    pthread_mutex_lock(&dec_mtx);
    *val = *val - 1;
    new = *val;
    pthread_mutex_unlock(&dec_mtx);
    return new;
}

#define NO_ATOMIC_COMPARE_AND_SWAP
#endif


/* A simple implementation of binary semaphores that spin before waiting */
typedef struct {
    unsigned int val;
    pthread_cond_t cond;
    pthread_mutex_t mtx;
} Sem_t;

static void Sem_init(Sem_t *s, unsigned int v)
{
    pthread_cond_init(&(s->cond), NULL);
    pthread_mutex_init(&(s->mtx), NULL);
    s->val = v;
}

static int Sem_check_and_reset(Sem_t *s)
{
#ifdef NO_ATOMIC_COMPARE_AND_SWAP
    /* to reduce contention this uses the semaphore's mutex rather
       than a single static mutex */
    int result = 0;
    pthread_mutex_lock(&(s->mtx));
    if (s->val) {
	result = 1;
	s->val = 0;
    }
    pthread_mutex_unlock(&(s->mtx));
    return result;
#else
    return atomic_compare_and_swap((int *) &(s->val), 1, 0);
#endif
}

static void Sem_wait(Sem_t *s, int adjust)
{
    long long count = spin_count > 0 ? spin_count : 1;

    /* spin while waiting */
    while (--count > 0 && ! Sem_check_and_reset(s));
    
    if (count == 0) {
	/* spinning completed without a post -- the last
	   compare_and_swap executed saw s->val == 0 */

	pthread_mutex_lock(&(s->mtx));
	/* we now have the mutex, so s->val cannot change until we
	   release the mutex */
	while (s->val == 0)
	    pthread_cond_wait(&(s->cond), &(s->mtx));
	s->val = 0;
	pthread_mutex_unlock(&(s->mtx));

	/* lower the spin rate if we ended up here; this should help
	   if we have more threads than available processors */
	if (adjust && spin_count > reduced_spin_count)
	    spin_count /= spin_reduce;
    }
}

static void Sem_post(Sem_t *s)
{
    pthread_mutex_lock(&(s->mtx));
    s->val = 1;
    pthread_mutex_unlock(&(s->mtx));
    /**** make sure it is safe to signal outside lock */
    pthread_cond_signal(&(s->cond));
}

static void Sem_destroy(Sem_t *s)
{
    pthread_cond_destroy(&(s->cond));
    pthread_mutex_destroy(&(s->mtx));
    s->val = 0;
}


/* Work crew implementation */
typedef struct { int index; } worker_info_t;

static Sem_t run_sem[MAX_NUM_WORKERS];
static Sem_t done_sem;
static pthread_t thread_id[MAX_NUM_WORKERS];
static worker_info_t worker_info[MAX_NUM_WORKERS];

static void *thread_routine(void *arg)
{
    worker_info_t *info = (worker_info_t *) arg;
    int index = info->index;

    while (1) {
	/* this needs to make sure the decrement is atomic and only
	   one thread sees numJobs hit zero and calls Sem_post */
	if (atomic_decrement(&numJobs) == 0)
	   Sem_post(&done_sem);
	Sem_wait(&run_sem[index], 0);
	work(index, data);
    }
    return NULL;
}

static void releaseWorkers(int njobs, void (*workfun)(int i, void *data),
			   void *workdata)
{
    int i;

    if (njobs > numWorkers)
	error("non enough workers available");

    numJobs = njobs;
    work = workfun;
    data = workdata;

    for (i = 0; i < njobs; i++)
	Sem_post(&run_sem[i]);
}

static void waitForWorkers()
{
    Sem_wait(&done_sem, 1);
}

static int joinWorkers()
{
    int i, status, join_failed = 0;
    void *thread_result;
    for (i = 0; i < numJobs; i++) {
	status = pthread_join(thread_id[i], &thread_result);
	if (status != 0) join_failed = 1;
	Sem_destroy(&run_sem[i]);
    }
    Sem_destroy(&done_sem);
    numJobs = 0;
    return join_failed;
}

static void exit_work(int i, void *data)
{
    pthread_exit(NULL);
}

static int startWorkerThreads(int n)
{
    int i, numStarted, status;

    numStarted = 0;
    Sem_init(&done_sem, 0);
    numJobs = n;
    for (i = 0; i < n; i++) {
	Sem_init(&run_sem[i], 0);
	worker_info[i].index = i;
	status = pthread_create(&thread_id[i], NULL, thread_routine,
				&worker_info[i]);
	if (status != 0) break;
	numStarted++;
    }
    waitForWorkers();
    return numStarted;
}
